/*
  zapojeni nozicek https://github.com/Xinyuan-LilyGO/TTGO-T-Display
  prace s prerusenim https://lastminuteengineers.com/handling-esp32-gpio-interrupts-tutorial/
*/

/*
 * Konfigurace vypoctu:
 * http://robodoupe.cz/2016/pozor-radiace-8/
 * http://robodoupe.cz/2016/pozor-radiace-9/
 * 
 * STS-5
 * 
 * 2500 imp/s = 30 uR
 * vlastni vyboj trubice max 27/min
 * 
 * 
 * SBM-20
 * 
 * 60 až 70 impulzů na 1 μR 
 * registrace β záření lepší než 40%
 * 1 mR/h velmi přibližně odpovídá 10 μSv/h
 * 
 * variantne
 * 
 * Citlivost γ (Ra226) 29 pulzů/s při 1 mR/hod
 * Citlivost γ (Co60)  22 pulzů/s při 1 mR/hod
 * Rozsah měření 0,014 až 144 mR/hod
 * 
 */


// +++ import pro displej
#include <TFT_eSPI.h>
#include <SPI.h>
#include <Wire.h>
#include "esp_adc_cal.h"
#include "bmp.h"
// --- import pro displej

// +++ data pro displej
#ifndef TFT_DISPOFF
#define TFT_DISPOFF 0x28
#endif

#ifndef TFT_SLPIN
#define TFT_SLPIN   0x10
#endif

#define TFT_MOSI            19
#define TFT_SCLK            18
#define TFT_CS              5
#define TFT_DC              16
#define TFT_RST             23

#define TFT_BL          4  // Display backlight control pin

TFT_eSPI tft = TFT_eSPI(135, 240); // Invoke custom library
// --- data pro displej

#include "src/ra/raLogger.h"
#include "RingBuffer.h"

raLogger* logger;

// buffer po data po jedne sekunde, drzime si 60 sekund
RingBuffer* rbPer1Sec;
// buffer pro data po 60 sec, drzime si 30 minut
RingBuffer* rbPer60Sec;

void IRAM_ATTR ISR();


#define INPUT_PIN 37

// jednou za 10 sec dumpovat do druheho bufgeru
int nactenychSekund;

// prubezne behem sekundy
volatile int pocetImpulsu;

// statisticka data
int impulzuZa1Sec;
int impulzuZa60Sec;
int impulzuCelkem;

int minPerMinute;
int maxPerMinute;

int tsReadData;
#define READ_PERIOD_MSEC 1000

#define MAX_MINUTOVYCH_ZAZNAMU 60

void setup() {
  Serial.begin(115200);

  logger = new raLogger();
  logger->log( "\n\n---start---" );

  rbPer1Sec = new RingBuffer(62);
  rbPer60Sec = new RingBuffer(MAX_MINUTOVYCH_ZAZNAMU);

  pocetImpulsu = 0;
  impulzuZa1Sec = 0;
  impulzuZa60Sec = 0;
  
  // konfigurace - toto je maximum udane datasheetem
  minPerMinute = -1; 
  maxPerMinute = 0;
  tsReadData = 0;
  nactenychSekund = 0;
  
  tft.init();
  
  pinMode( INPUT_PIN, INPUT_PULLUP );
  attachInterrupt( INPUT_PIN, ISR, RISING);
}


// je volano v preruseni
void IRAM_ATTR ISR() {
    pocetImpulsu++;
}

// pocet nevyplnenych hodnot za dany soucet
int ct_invalid;

/*
 * Side efect! Nastavuje ct_invalid
 */
int sectiObsah1SecBufferu( int ct )
{
    int suma = 0;
    ct_invalid = 0;
    
    for( int i = 0; i<ct; i++ ) {
        int v = rbPer1Sec->get(i);
        if( v!=RB_NODATA ) {
          suma += v;  
        } else {
          ct_invalid++;
        }
    }
    return suma;
}


/**
 * Side efect: nastavuje minPerMinute, maxPerMinute
 */
void nactiData()
{
  noInterrupts();
  impulzuZa1Sec = pocetImpulsu;
  pocetImpulsu = 0;
  interrupts();

  // jednou za 60 sec spocteme sumu a vlozime ji do per60sec bufferu
  nactenychSekund++;
  if( nactenychSekund==60 ) {
     rbPer60Sec->put( sectiObsah1SecBufferu(60) ); 
     nactenychSekund = 0;
  }
  
  rbPer1Sec->put( impulzuZa1Sec );

  // ulozime do globalni promenne, bude se pouzivat
  impulzuZa60Sec = sectiObsah1SecBufferu(60);

  if( ct_invalid==0 ) {
     if( minPerMinute==-1 ) {
        minPerMinute = impulzuZa60Sec;
     }
     if( impulzuZa60Sec < minPerMinute ) {
        minPerMinute = impulzuZa60Sec;
     }
     if( impulzuZa60Sec > maxPerMinute ) {
        maxPerMinute = impulzuZa60Sec;
     }
  }

  logger->log( "last60s=%d lastmin=%d", 
      impulzuZa60Sec,
      rbPer60Sec->get( 0 )
      );
  rbPer60Sec->dumpContent( logger, 10 );
}


/*
 * prepocet poctu impulzu za cas na R/h resp. Sv/h
 */
double spoctiIntenzitu( int pocetImpulzu, int dobaMin )
{
    // suma impulsu za zadany cas
    // prepoctovy koeficient 
    //        gama: 60 až 70 impulzů na 1 μR 
    //        beta: min 40 %
    // 1 mR/h velmi přibližně odpovídá 10 μSv/h.  .... 1000 uR = 10 uS

#define IMP_PER_UR 65
#define UR_TO_USV 100

    double uR = (double)pocetImpulzu / (double)IMP_PER_UR;
    double hour = (double)dobaMin / (double)60;
    double uR_per_hour = uR / hour;
    double uSv_per_hour = uR_per_hour / UR_TO_USV;

    logger->log( "%d/%d -> uR %.1f, hr %.1f, uR/h %.1f, uSv/h %.1f",
      pocetImpulzu,
      dobaMin,
      uR,
      hour,
      uR_per_hour,
      uSv_per_hour );

    return uSv_per_hour;
}


int avg;
double avgIntenzita;

// vlastni vyboje trubice - vstupni hodnota, pokud jsme nezmerili mene
#define VLASTNI_VYBOJE_TRUBICE 23

// vlastni vyboje trubice - hodnota pouzivana k vypoctum
int base;

void spoctiData()
{
    int suma = 0;
    int pocet = 0;
    
    for( int i = 0; i<MAX_MINUTOVYCH_ZAZNAMU; i++ ) {
        int v = rbPer60Sec->get(i);
        if( v==RB_NODATA ) {
          break;
        } 
        suma += v;
        pocet++;
    }

    if( pocet==0 ) {
      avg = -1;
      return;
    }

    suma = suma - (pocet*base);
    avg =  suma / pocet;

    logger->log( "prumerMinut: suma %d, pocet %d, avg %d", 
      suma,
      pocet, 
      avg
      );

    avgIntenzita = spoctiIntenzitu( suma, pocet );

    return;
}

// s odectenym base
int per60sec;
// bez odecteneho base - pro graf
int per60secGraph;


// pocet impulzu v aktualni minute, 
// pokud je namerena mene nez minuta, extrapoluje do minuty!
// uklada do globalni promenne per60sec - s odectenym zakladem (vlastnim zarenim)
// a do per60secGraph - bez zakladu
void spoctiAktualniMinutu()
{
    per60secGraph = impulzuZa60Sec; 
    
    if( millis()<60000 ) {
        double pomer = (double)millis() / 60000.0;
        double num = (double)per60secGraph / pomer;
        per60secGraph = (int)num;
    }
    per60sec = per60secGraph - base;
    if( per60sec<0 ) { per60sec = 0; }  
}


/**
 * Naformatovani intenzity do stringu, vcetne jednotek.
 */
void formatujSv( char * buff, double val )
{
    if( val > 1000.0 ) {
        sprintf( buff, "%.2f mSv/h", (val/1000.0) );
    } else if( val > 100.0 ) {
        sprintf( buff, "%d uSv/h", (int)val );
    } else if( val > 10.0 ) {
        sprintf( buff, "%.1f uSv/h", val );
    } else {
        sprintf( buff, "%.2f uSv/h", val );
    } 
}


#define X_COL_1 0
#define X_COL_2 100
#define X_OFF_NADPIS 5

#define NAPIS_MALY_VYSKA 12
#define NAPIS_VELKY_VYSKA 20


void tftPrint( int color, int textSize,
            int pos_x,
            int pos_y_malych_napisu, int pos_y_velkych_napisu,
            char * buff )
{
    tft.setTextColor(color, TFT_BLACK);
    tft.setTextSize(textSize);
    int pos_y = pos_y_malych_napisu*NAPIS_MALY_VYSKA + pos_y_velkych_napisu*NAPIS_VELKY_VYSKA;
    tft.setCursor( pos_x, pos_y );
    tft.print( buff );
}

// spocte vlastni vyboje trubice - zaklad pro dalsi vypocty
void spoctiBase()
{
    // kolik impulzu za minutu odecteme
    base = minPerMinute;
    if( base==-1 || base>VLASTNI_VYBOJE_TRUBICE ) {
       base = VLASTNI_VYBOJE_TRUBICE;
    }
    logger->log( "base %d (min-max %d %d)", base, minPerMinute, maxPerMinute );    
}


/*
 * Vypsani namerenych hodnot
 */
void displayData()
{
    tft.setRotation(3);
    tft.fillScreen(TFT_BLACK);
    tft.setTextDatum(TL_DATUM); // kresli se od top-left

    char buff[512];

    int curTimeSec = millis() / 1000;
    int curMin = curTimeSec / 60;
    int curSec = curTimeSec % 60;

    // spocte hodnoty
    spoctiBase();
    spoctiData();
    spoctiAktualniMinutu();

    double intenzitaNow = spoctiIntenzitu( per60sec, 1 );


  // cas
    tftPrint( TFT_WHITE, 1,
            X_COL_1,
            0, 0,
            "time:" );

    sprintf( buff, "%d:%02d",
          curMin,
          curSec );
          
    tftPrint( TFT_GREEN, 2,
            X_COL_1+X_OFF_NADPIS,
            1, 0,
            buff );

  
  // pocet impulsu/sec za posledni plovouci minutu


    tftPrint( TFT_WHITE, 1,
            X_COL_1,
            1, 1,
            "1/min:" );

    sprintf( buff, "%d", per60sec );
          
    tftPrint( TFT_GREEN, 2,
            X_COL_1+X_OFF_NADPIS,
            2, 1,
            buff );
            


  // pocet impulsu/min prumer
    if( avg>0 ) {

        tftPrint( TFT_WHITE, 1,
                X_COL_1,
                2, 2,
                "1/min avg:" );
    
        sprintf( buff, "%d", avg );
              
        tftPrint( TFT_GREEN, 2,
                X_COL_1+X_OFF_NADPIS,
                3, 2,
                buff );
    }

    // min/max za minutu
    if( minPerMinute!=-1 ) {

        // rotujeme hodnoty min / max po 4 sec
        bool showMin = false;
        if( nactenychSekund & 4 ) {
            showMin = true;
        }

        tftPrint( TFT_WHITE, 1,
                X_COL_1,
                3, 3,
                showMin ? (char*)"min:" : (char*)"max:" );

        if( showMin ) {
            sprintf( buff, "%d..", minPerMinute );
        } else {
            sprintf( buff, "..%d", maxPerMinute );
        }
        
        tftPrint( TFT_GREEN, 2,
                X_COL_1+X_OFF_NADPIS,
                4, 3,
                buff );

    }

    // intenzita nyni
    if( intenzitaNow > 0.0 ) {

      tftPrint( TFT_WHITE, 1,
            X_COL_2,
            0, 0,
            "intensity:" );

      formatujSv( buff, intenzitaNow );
            
      tftPrint( TFT_GREEN, 2,
              X_COL_2+X_OFF_NADPIS,
              1, 0,
              buff );
    }

    // intenzita prumerna
    if( avg > 0 ) {

      tftPrint( TFT_WHITE, 1,
            X_COL_2,
            1, 1,
            "intensity avg:" );

      formatujSv( buff, avgIntenzita );
            
      tftPrint( TFT_GREEN, 2,
              X_COL_2+X_OFF_NADPIS,
              2, 1,
              buff );
    }
}



#define GRAPH_X X_COL_2+X_OFF_NADPIS
#define GRAPH_Y 133
#define GRAPH_SIZE_Y 60.0
#define GRAPH_BAR_WIDTH 4
#define GRAPH_BAR_WIDTH_FIRST 6
#define GRAPH_VALS 30


/*
 * Vykresleni grafu
 */
void displayGraph()
{
      /* int32_t x, int32_t y, int32_t h, uint32_t color
        x=0,y=0 = levy horni roh
        souradnice se zvetsuji doprava a dolu
        size 135, 240
        */

    int max=10;
    int min=999999999;
    for( int i = 0; i<GRAPH_VALS; i++ ) {
       int v = rbPer60Sec->get( i );
       if( v==RB_NODATA ) break;
       if( v>max ) max=v;
       if( v<min ) min=v;
    }

    if( per60secGraph>max ) max=per60secGraph;
    if( per60secGraph<min ) min=per60secGraph;

    // zvetsime rozsah min-max, aby to nebylo tak citlive pri malych tocich (pri vysokych je to zanedbatelne)
    min = min - 5;
    max = max + 20;

    float range = max-min;
    if( range==0 ) range=1;
    range = GRAPH_SIZE_Y / range;

    logger->log( "graph: %d-%d, %f, per60=%d", min, max, range, per60secGraph );

    int x_pos = GRAPH_X;

    // vykreslit aktualni minutu
    float vd = ((float)(per60secGraph-min)) * range;
    int len = (int)vd + 2;
    for(int j=0; j<(GRAPH_BAR_WIDTH_FIRST); j++ ) {
      tft.drawFastVLine( x_pos, GRAPH_Y-len, len, TFT_RED ); 
      x_pos++;      
    }

    // vykreslit dalsi historii
    for( int i = 0; i<GRAPH_VALS; i++ ) {
       int v = rbPer60Sec->get( i );
       if( v==RB_NODATA ) break;

       float vd = ((float)(v-min)) * range;
       int len = (int)vd + 2;

       for(int j=0; j<GRAPH_BAR_WIDTH; j++ ) {
          tft.drawFastVLine( x_pos, GRAPH_Y-len, len, TFT_GREEN ); 
          x_pos++;      
       }
    }
}


// drawFastVLine(int32_t x, int32_t y, int32_t h, uint32_t color),

void loop() {

  int currentMillis = millis();  
  if (currentMillis - tsReadData >= READ_PERIOD_MSEC) 
  {
    nactiData();
    tsReadData = currentMillis;  
    displayData();
    if( millis()>120000) {
        displayGraph();
    } else {
      tftPrint( TFT_WHITE, 1,
            X_COL_2,
            3, 3,
            "Graf za chvili..." );
    }
  }

}
